/****************************************************************************
**
** Copyright (C) 2006-2009 fullmetalcoder <fullmetalcoder@hotmail.fr>
**
** This file is part of the Edyuk project <http://edyuk.org>
** 
** This file may be used under the terms of the GNU General Public License
** version 3 as published by the Free Software Foundation and appearing in the
** file GPL.txt included in the packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/

#include "qnfadefinition.h"

/*!
	\file qnfadefinition.cpp
	\brief Implementation of the QNFADefinition class.
*/

#include "qnfa.h"

#include "qformatscheme.h"

#include "qmarksmanager.h"

#include "qdocument.h"
#include "qdocument_p.h"
#include "qdocumentline.h"
#include "qdocumentcursor.h"

#include <QDebug>

#include <QFile>
#include <QKeyEvent>

#include <QDomText>
#include <QDomElement>
#include <QDomDocument>

uint qHash(const QPointer<QDocument>& p) { return qHash((QDocument*)(p)); }

extern QString *_singleLineCommentTarget;
void embed(QNFA *src, QNFA *dest, int index);

extern bool stringToBool(const QString& s, bool previous);

QList<QNFADefinition::Token> QNFADefinition::m_tokens;
QList<QNFADefinition::Paren> QNFADefinition::m_parentheses;
QHash<int, QNFADefinition::Context> QNFADefinition::m_contexts;
QHash<QNFA*, int> QNFADefinition::m_contextIds;
QHash<QString, int> QNFADefinition::m_paren;
QHash<QString, QNFA*> QNFADefinition::m_contextNames;
QHash<QString, QNFADefinition::EmbedRequestList> QNFADefinition::m_pendingEmbeds;

void QNFADefinition::load(const QString& file, QLanguageFactory::LangData *d, QFormatScheme *s)
{
	QFile f(file);
	
	if ( !f.open(QFile::ReadOnly | QFile::Text) )
	{
		qWarning("QNFADefinition : failed to open file %s", qPrintable(file));
		return;
	}
	
	load(&f, d, s);
}

void QNFADefinition::load(QFile *f, QLanguageFactory::LangData *d, QFormatScheme *s)
{
	QDomDocument doc;
	doc.setContent(f);
	
	load(doc, d, s);
}

void QNFADefinition::load(const QDomDocument& doc, QLanguageFactory::LangData *d, QFormatScheme *s)
{
	QDomElement root = doc.documentElement();
	
	QNFADefinition *nd = new QNFADefinition;
	
	nd->m_language = d->lang = root.attribute("language");
	nd->m_indentFold = stringToBool(root.attribute("indentationFold"), false);
	nd->m_extensions = d->extensions = root.attribute("extensions").split(";");
	nd->m_defaultMark = root.attribute("defaultLineMark", "bookmark");
	
	/*
	qDebug("Generating definition for %s language from XML file : %s",
			qPrintable(m_language),
			qPrintable(fn));
	*/
	
	// create root entity
	nd->m_root = lexer();
	
	_singleLineCommentTarget = &(nd->m_singleLineComment);
	nd->fillContext(nd->m_root, root, s, true);
	_singleLineCommentTarget = 0;
	
	squeeze(nd->m_root);
	
	//m_definitions[m_language] = this;
	m_contextNames[nd->m_language] = nd->m_root;
	
	flushEmbedRequests(nd->m_language);
	
	d->d = nd;
	d->e = 0;
	d->s = s;
}

QNFADefinition::QNFADefinition()
 : m_indentFold(false), m_root(0)
{
	
}

QNFADefinition::~QNFADefinition()
{
	delete m_root;
}

QString QNFADefinition::language() const
{
	return m_language;
}

QStringList QNFADefinition::extensions() const
{
	return m_extensions;
}

int QNFADefinition::tokenFormat(int tokenId) const
{
	//qDebug("%i::%i", tokenId & TokenIdMask, m_tokens.at(tokenId & TokenIdMask).format);
	return m_tokens.at(tokenId & TokenIdMask).format;
}

int QNFADefinition::contextFormat(int contextId) const
{
	//qDebug("%i::%i", contextId & ContextIdMask, m_contexts.value(contextId & ContextIdMask).format);
	return m_contexts.value(contextId & ContextIdMask).format;
}

inline bool isWord(QChar c)
{ return c.isLetterOrNumber() || (c.unicode() == '_'); }

static inline bool match(QChar cc, const QNFA *chain)
{
	const quint16 cu = cc.unicode();
	const int n = chain->c.count();
	const int ass = chain->assertion;
	quint16 fc = n ? chain->c.at(0) : 0;
	
	bool found = !(n && !fc);
	
	// avoid conditionals at all cost...
	
	const bool fallback = chain->c.contains(cu) ? found : !found;
	const bool isDigit = cc.isDigit();
	const bool isSpace = cc.isSpace();
	const bool isWord = cc.isLetterOrNumber() || cu == '_';
	
	return
			((n == 1) && (fc == cu))
		?
			true
		:
			(
				ass
			?
				(
					(
						((ass & Digit) && isDigit)
					||
						((ass & Space) && isSpace)
					||
						((ass & Word) && isWord)
					||
						((ass & NonDigit) && !isDigit)
					||
						((ass & NonSpace) && !isSpace)
					||
						((ass & NonWord) && !isWord)
					)
				?
					found
				:
					fallback
				)
			:
				fallback
			)
		;
}

void QNFADefinition::tokenize(
								QDocumentLineHandle *h,
								QNFA*& context, QStack<QNFA*>& cxt,
								QStack<int>& cxtidx, QMap<int, QPair<int, int> >& cxtData
							)
{
	int offset, length = h->length();
	const QChar *d = h->text().constData();
	
	quint16 c = 0;
	const QChar *di = d;
	QNFA *chain = 0, *start = 0;
	int index = 0, lastCxt = 0, len;
	bool bFound, bEscape = false, bEscaped = false;
	bool wPrev = false, wCur = false;
	
	while ( index <= length )
	{
		bFound = false;
		bEscaped = false;
		
		if ( index == length )
		{
			c = '\n';
			wCur = false;
		} else {
			c = di->unicode();
			wCur = isWord(*di);
		}
		
		int plainIndex = -1, plainMatch, plainLength;
		
		// try fast plain matching
		if ( !(wPrev && wCur) )
		{
			//qDebug("trying plain...");
			
			//len = 0;
			int idx = index;
			QCharTree::const_iterator it, match, end;
			
			it = context->tree.constFind(c);
			
			if ( it != context->tree.constEnd() )
			{
				//qDebug("plain on %c", c);
				do
				{
					++di;
					++idx;
					
					end = it->next.constEnd();
					match = it->next.constFind(0);
					
					if ( idx <= length )
					{
						c = idx == length ? '\n' : di->unicode();
						it = it->next.constFind(c);
					} else {
						it = end;
					}
					
					if ( it == end )
					{
						if ( (match != end) && !isWord(*di) )
						{
							//word boundary found
							// corresponding token end found
							wPrev = isWord(*(di - 1));
							bFound = true;
							if ( match->value.action & 0x40000000 )
							{
								// try regexps before notifying
								plainIndex = index;
								plainLength = idx - index;
								plainMatch = match->value.action;
								//qDebug("ambiguity.");
							} else {
								h->addToken(index, idx - index, match->value.action);
								index = idx;
							}
							//qDebug("next step : %c", d[index].toLatin1());
							//bMonitor = true;
						}
						
						break;
					}
				} while ( idx < length );
				
				if ( bFound )
				{
					bEscape = false;
					
					if ( plainIndex == -1 )
						continue;
					
					bFound = false;
				}
				
				di -= idx - index;
			}
		}
		
		// fallback on regexp-like NFA-based matching
		QNFABranch* children = context->out.branch;
		
		if ( children )
		{
			//qDebug("trying %i sub nfas on %c", children->count(), d[index].toLatin1());
			const int max = children->count();
			
			for ( quint16 i = 0; i < max; ++i )
			{
				len = 0;
				int idx = index;
				start = chain = children->at(i);
				
				//qDebug("%ith attempt on %c", i, d[index + len].toLatin1());
				
				while ( (idx <= length) || (chain->type & Match) )
				{
					bEscaped = false;
					QChar cc = idx < length ? *di : '\n';
					
					if ( chain->type & Match )
					{
						if (
								(chain->assertion & WordEnd)
							&&
								isWord(*di)
							&&
								isWord(*(di - 1))
							)
						{
							//qDebug("end assertion failed...");
							break;
						}
						
						//qDebug("matched to end");
						
						if ( chain->type & CxtBeg )
						{
							//qDebug("entering context : 0x%x", chain);
							
							h->addToken(index, len, start->actionid ? start->actionid : chain->actionid);
							
							// update context stack
							cxt.push(context);
							
							context = chain;
							cxtidx.push(idx);
							
						} else if ( chain->type & CxtEnd ) {
							//qDebug("leaving context :");
							
							if ( cxt.isEmpty() )
								qFatal("context nesting problem");
							
							if ( bEscape )
							{
								// not really end : escape found...
								
								bEscape = false;
								bEscaped = true;
							} else {
								int cxtlen;
								
								if ( cxtidx.count() )
								{
									lastCxt = cxtidx.pop();
									cxtlen = index - lastCxt;
								} else {
									lastCxt = -1;
									cxtlen = index;
								}
								
								h->addToken(index, len, chain->actionid);
								
								int cid = m_contextIds.value(context);
								//if ( lastCxt != -1 || !(context->type & StayOnLine) )
								if ( lastCxt == -1 )
								{
									h->addContext(-1, cxtlen, cid);
								} else {
									cxtData.insertMulti(lastCxt, qMakePair(cxtlen, cid));
								}
								
								// update context stack
								context = cxt.pop();
								
								//qDebug("cxt notif...");
								
								if ( chain->type & Exclusive )
								{
									index = idx;
								} else {
									//qDebug("non-exclusive");
									di -= idx - index;
								}
								
								--index;
								--di;
								
								bFound = true;
								break;
							}
						} else if ( chain->type & CxtEsc )  {
							//qDebug("matched %s", qPrintable(QString(index, len)));
							
							//notify(index, len, chain->actionid);
							bEscape = !bEscape;
						} else {
							//qDebug("matched %s", qPrintable(QString(d + index, len)));
							
							if ( plainIndex != -1 && plainLength >= len )
							{
								break;
							}
							
							//notify(index, len, chain->actionid);
							h->addToken(index, len, chain->actionid);
							bEscape = false;
						}
						
						bFound = true;
						index = idx;
						--index;
						--di;
						
						//qDebug("next step : %c", d[index + 1].toLatin1());
						//bMonitor = true;
						
						break;
					} else {
						// "regular" nfa match (no match yet...)
						
						if (
								(chain->assertion & WordStart)
							&&
								(idx >= 1)
							&&
								(
									isWord(*(di - 1))
								&&
									isWord(cc)
								)
							)
						{
							//qDebug("beg assertion failed...");
							
							break;
						}
						
						bool found = ::match(cc, chain);
						
						if (
								!(chain->assertion & ZeroOrOne)
							&&
								!(chain->assertion & ZeroOrMore)
							&&
								!found
							)
						{
							//if ( cc.toLatin1() == ')' )
							//	qDebug("mismatch : %c != %c", cc.toLatin1(), chain->c.at(0));
							
							break;
						}
						
						if ( found )
						{
							//qDebug("%c", d[index + len].toLatin1());
							
							if (
									(chain->assertion & OneOrMore)
								||
									(chain->assertion & ZeroOrMore)
								)
							{
								do
								{
									++di;
									++len;
									++idx;
									
									if ( idx > length )
										break;
									
									if ( idx == length )
										cc = '\n';
									else
										cc = *di;
									
								} while ( ::match(cc, chain) );
								
							} else {
								++len;
								++idx;
								++di;
							}
							
						} else {
							//qDebug("! %c", d[index + len].toLatin1());
						}
						
						chain = chain->out.next;
					}
				}
				
				if ( bFound )
					break;
				
				di -= len;
			}
		}
		
		if ( !bFound )
		{
			if ( plainIndex != -1 )
			{
				//notify(plainIndex, plainLength, plainMatch);
				h->addToken(plainIndex, plainLength, plainMatch);
				index = plainIndex + plainLength;
				di += plainLength;
				continue;
			}
			
			bEscape = false;
			//++index;
			wPrev = wCur;
		} else if ( index < length ) {
			wPrev = isWord(*di);
		}
		
		++index;
		++di;
	}
}

/*!
	\brief Entry point for syntax highlighting
*/
int QNFADefinition::tokenize(QDocument *d, int line, int count)
{
	QNFA *context;
	QStack<QNFA*> cxt;
	const int startLine = line, lastLine = line + count;
	QDocumentBuffer::iterator it = d->impl()->find(line);
	
	cxt.push(m_root);
	
	if ( line > 0 )
	{
		int min = 0;
		QDocumentBuffer::iterator tmp = it;
		tmp.moveBy(-1);
		
		const QVector<int>& cxts = tmp.lineHandle()->contexts();
		
		//qDebug() << cxts;
		
		for ( int i = min; i + 2 < cxts.count(); i += 3 )
		{
			// at this point we're only interested in context going forth
			if ( cxts.at(i + 1) != -1 )
				continue;
			
			int cxtId = cxts.at(i + 2);
			
			// language are just there for proper context awareness, they
			// serve no actual purpose in the context stack of the lexer...
			if ( cxtId & Language )
				continue;
			
			//cxt.push(m_contextNames.value(contextName(cxtId)));
			cxt.push(m_contexts.value(cxtId).nfa);
		}
	}
	
	//qDebug() << cxt;
	
	context = cxt.pop();
	
	QDocumentLineHandle *h = 0;
	
	// lexing loop, line by line
	while ( !it.atEnd() )
	{
		h = it.lineHandle();
		
		QVector<int> oldCxtStack, newCxtStack;
		
		if ( line + 1 >= lastLine )
		{
			const QVector<int>& cxts = h->contexts();
			
			for ( int i = 0; i + 2 < cxts.count(); i += 3 )
			{
				if ( cxts.at(i + 1) == -1 )
				{
					int cxtId = cxts.at(i + 2);
					
					oldCxtStack.append(cxtId);
				}
			}
		}
		
		h->clearTokens();
		h->clearContexts();
		
		QStack<int> cxtidx;
		QMap<int, QPair<int, int> > cxtData;
		
		tokenize(h, context, cxt, cxtidx, cxtData);
		
		++line;
		
		//qDebug("%i, %i", line, lastLine);
		
		cxt.push(context);
		
		// commit contexts in proper order
		const int n = cxt.count(), nn = cxtidx.count();
		const int old = n - nn;
		
		QList<int> stay;
		
		for ( int i = 1; i < n; ++i )
		{
			int cxtId = m_contextIds.value(cxt.at(i));
			
			if ( i < (n - 1) && (cxt.at(i)->type & StayOnLine) )
			{
				//qDebug("stay...");
				stay << i;
				
				if ( i < old )
					h->addContext(-1, h->length(), cxtId);
				else
					cxtData.insertMulti(cxtidx.at(i - old), qMakePair(h->length() - (i - old), cxtId));
				
				continue;
			}
			
			//qDebug("%i:%i", i, cxtId);
			
			if ( i < old )
			{
				//qDebug(">][ : %i", cxtId);
				h->addContext(-1, -1, cxtId);
				
				if ( line >= lastLine )
					newCxtStack.append(cxtId);
				
			} else {
				//qDebug(">[%i,[ : %i", cxtidx.at(k), cxtId);
				cxtData.insertMulti(cxtidx.at(i - old), qMakePair(-1, cxtId));
			}
		}
		
		cxt.pop();
		
		for ( int i = stay.count() - 1; i >= 0; --i )
			cxt.remove(stay.at(i));
		
		QMap<int, QPair<int, int> >::const_iterator cit = cxtData.constBegin();
		
		while ( cit != cxtData.constEnd() )
		{
			//qDebug(">[%i,%i? : %i", cit.key(), cit->first, cit->second);
			h->addContext(cit.key(), cit->first, cit->second);
			
			if ( (cit->first == -1) && (line >= lastLine) )
				newCxtStack.append(cit->second);
			
			++cit;
		}
		
		it.moveBy(1);
		
		//qDebug() << line << cxt;
		
		if ( line >= lastLine )
		{
			// check whether further lexing is required
			//qWarning("TODO : continuation");
			
			//qDebug() << oldCxtStack << " vs " << newCxtStack;
			
			if ( newCxtStack == oldCxtStack )
				break;
			
		}
	}
	
	return line - startLine;
}

/*!
	\brief Return the string starting a single line comment, if any offered by the language
*/
QString QNFADefinition::singleLineComment() const
{
	return m_singleLineComment;
}

QString QNFADefinition::defaultLineMark() const
{
	return m_defaultMark;
}

/*!
	\brief Brace matching entry point
*/
void QNFADefinition::clearMatches(QDocument *d)
{
	QHash<QPointer<QDocument>, int>::iterator it = m_matchGroups.find(d);
	
	if ( it != m_matchGroups.end() )
	{
		d->clearMatches(*it);
		d->flushMatches(*it);
		
		// erase?
		//m_matchGroups.erase(it);
	}
}

/*!
	\brief Brace matching entry point
*/
void QNFADefinition::match(QDocumentCursor& c)
{
	QDocumentLine b = c.line();
	QDocument *d = c.document();
	
	if ( !d || !b.isValid() )
	{
		qDebug("invalid cursor");
		return;
	}
	
	int gid = 0;
	QHash<QPointer<QDocument>, int>::iterator it = m_matchGroups.find(d);
	
	if ( it != m_matchGroups.end() )
	{
		d->clearMatches(*it);
		d->flushMatches(*it);
		*it = gid = d->getNextGroupId();
	} else {
		gid = d->getNextGroupId();
		m_matchGroups[d] = gid;
	}
	
	QFormatScheme *s = d->formatScheme();
	
	if ( !s )
		s = QDocument::defaultFormatScheme();
	
	if ( !s )
	{
		qDebug("no fmt");
		d->releaseGroupId(gid);
		m_matchGroups.remove(d);
		return;
	}
	
	const int matchFID = s->id("braceMatch"), mismatchFID = s->id("braceMismatch");
	
	int matchCount = 0;
	int pos = c.columnNumber();
	
	const QVector<int>& m_tokens = b.tokens();
	
	if ( m_tokens.count() < 4 )
	{
		// required to properly update display
		d->releaseGroupId(gid);
		m_matchGroups.remove(d);
		return;
	}
	
	//qDebug("matching on line %i (fid is %i)", c.lineNumber(), fid);
	
	int tokenIndex = 1;
	int tokenCount = m_tokens.first();
	
	while ( tokenIndex + 2 < tokenCount )
	{
		int offset = m_tokens.at(tokenIndex);
		int length = m_tokens.at(++tokenIndex);
		int tokenId = m_tokens.at(++tokenIndex);
		++tokenIndex;
		
		if (
				(pos != offset)
			&&
				(pos != (offset + length))
			&&
				!(tokenId & QLanguageDefinition::Paren)
			&&
				!(tokenId & QLanguageDefinition::BraceMatch)
			)
			continue;
		
		PMatch m;
		m.line[0] = c.lineNumber();
		m.column[0] = offset;
		m.length[0] = length;
		
		if ( (tokenId & QLanguageDefinition::OpenParen) && (tokenId & QLanguageDefinition::CloseParen) )
		{
			//qDebug("match bound");
			
			matchOpen(d, m);
			
			if ( m.type == PMatch::Match )
			{
				++matchCount;
				
				d->addMatch(gid, m.line[0], m.column[0], m.length[0], matchFID);
				d->addMatch(gid, m.line[1], m.column[1], m.length[1], matchFID);
			}
			
			m.line[0] = c.lineNumber();
			m.column[0] = offset;
			m.length[0] = length;

			matchClose(d, m);
			
			if ( m.type == PMatch::Match )
			{
				++matchCount;
				
				d->addMatch(gid, m.line[0], m.column[0], m.length[0], matchFID);
				d->addMatch(gid, m.line[1], m.column[1], m.length[1], matchFID);
			}
			
			continue;
		} else if ( tokenId & QLanguageDefinition::OpenParen ) {
			//qDebug("match open");
			
			matchOpen(d, m);
		} else if ( tokenId & QLanguageDefinition::CloseParen ) {
			//qDebug("match close");
			
			matchClose(d, m);
		}
		
		if ( m.type != PMatch::Invalid )
		{
			++matchCount;
			
			int mfid = m.type == PMatch::Match ? matchFID : mismatchFID;
			d->addMatch(gid, m.line[0], m.column[0], m.length[0], mfid);
			d->addMatch(gid, m.line[1], m.column[1], m.length[1], mfid);
		}
	}
	
	if ( matchCount )
	{
		d->flushMatches(gid);
	} else {
		//qDebug("match failed");
		
		d->releaseGroupId(gid);
		m_matchGroups.remove(d);
	}
}

void QNFADefinition::matchOpen(QDocument *d, PMatch& m)
{
	int line = m.line[0];
	int pos = m.column[0];
	QVector<int> tokens = d->line(line).tokens();
	
	bool bMatch = false;
	int offset, length;
	QStack<int> lines;
	QStack<bool> type;
	QStack<int> parens;
	QStack<int> columns;
	QStack<int> lengths;
	
	forever
	{
		if ( tokens.count() > 3 )
		{
		int tokenIndex = 1;
		int tokenCount = tokens.first();
		
		while ( tokenIndex + 2 < tokenCount )
		{
			offset = tokens.at(tokenIndex);
			length = tokens.at(++tokenIndex);
			int tokenId = tokens.at(++tokenIndex);
			++tokenIndex;
			
			if ( !(tokenId & QLanguageDefinition::Paren) || !(tokenId & QLanguageDefinition::BraceMatch) )
				continue;
			
			bool open = tokenId & QLanguageDefinition::OpenParen;
			bool close = tokenId & QLanguageDefinition::CloseParen;
			
			if ( !(open || close) || offset < pos )
				continue;
			
			int parenId = m_tokens.at(tokenId & QLanguageDefinition::TokenIdMask).parenId;
			
			if ( open && close )
			{
				bool bInsert = true;
				
				if ( parens.count() )
				{
					if ( parens.top() == parenId )
					{
						//qDebug("submatch at %i", line);
						type.pop();
						lines.pop();
						parens.pop();
						columns.pop();
						lengths.pop();
						
						if ( parens.isEmpty() )
						{
							bMatch = true;
							//qDebug("match at %i", line);
							m.type = PMatch::Match;
							
							break;
						}
					} else {
						bInsert = false;
						
						do
						{
							int tp = parens.top();
							
							if ( type.top() )
							{
								if ( parenId < tp )
								{
									type.pop();
									lines.pop();
									parens.pop();
									columns.pop();
									lengths.pop();
								} else {
									break;
								}
							} else {
								break;
							}
							
						} while ( parens.count() );
						
						if ( parens.isEmpty() )
						{
							bMatch = true;
							m.type = PMatch::Mismatch;
							break;
						}
					}
				}
				
				if ( bInsert )
				{
					parens.push(parenId);
					columns.push(offset);
					lengths.push(length);
					type.push(true);
					lines.push(line);
				}
			} else if ( open ) {
				lines.push(line);
				type.push(false);
				columns.push(offset);
				parens.push(parenId);
				lengths.push(length);
			} else if ( close ) {
				if ( (bMatch = (parens.top() == parenId) ) )
				{
					bMatch &= parens.count() == 1;
					
					if ( bMatch )
					{
						m.type = PMatch::Match;
						
						break;
					} else if ( parens.count() ) {
						parens.pop();
						lines.pop();
						type.pop();
						lengths.pop();
						columns.pop();
					} else
						qWarning("bad paren nesting...");
				} else {
					bMatch = true;
					
					parens.pop();
					type.pop();
					m.line[0] = lines.pop();
					m.column[0] = columns.pop();
					m.length[0] = lengths.pop();
					
					m.type = PMatch::Mismatch;
					
					break;
				}
			}
		}
		
		if ( bMatch )
			break;
		}
		
		pos = 0;
		++line;
		
		QDocumentLine b = d->line(line);
		
		if ( !b.isValid() )
		{
			m.type = PMatch::Invalid;
			return;
		}
		
		tokens = b.tokens();
	}
	
	m.line[1] = line;
	m.column[1] = offset;
	m.length[1] = length;
}

void QNFADefinition::matchClose(QDocument *d, PMatch& m)
{
	int line = m.line[0];
	int pos = m.column[0];
	QVector<int> tokens = d->line(line).tokens();
	
	bool bMatch = false;
	int offset, length;
	QStack<int> lines;
	QStack<bool> type;
	QStack<int> parens;
	QStack<int> columns;
	QStack<int> lengths;
	
	forever
	{
		if ( tokens.count() > 3 )
		{
		
		int tokenIndex = tokens.first();
		
		while ( tokenIndex > 3 )
		{
			int tokenId = tokens.at(--tokenIndex);
			length = tokens.at(--tokenIndex);
			offset = tokens.at(--tokenIndex);
			
			if ( !(tokenId & QLanguageDefinition::Paren) || !(tokenId & QLanguageDefinition::BraceMatch) )
				continue;
			
			bool open = tokenId & QLanguageDefinition::OpenParen;
			bool close = tokenId & QLanguageDefinition::CloseParen;
			
			if ( !(open || close) || offset > pos )
				continue;
			
			int parenId = m_tokens.at(tokenId & QLanguageDefinition::TokenIdMask).parenId;
			
			if ( open && close )
			{
				bool bInsert = true;
				
				if ( parens.count() )
				{
					if ( parens.top() == parenId )
					{
						//qDebug("submatch at %i", line);
						type.pop();
						lines.pop();
						parens.pop();
						columns.pop();
						lengths.pop();
						
						if ( parens.isEmpty() )
						{
							bMatch = true;
							//qDebug("match at %i", line);
							m.type = PMatch::Match;
							
							break;
						}
					} else {
						bInsert = false;
						
						do
						{
							int tp = parens.top();
							
							if ( type.top() )
							{
								if ( parenId < tp )
								{
									type.pop();
									lines.pop();
									parens.pop();
									columns.pop();
									lengths.pop();
								} else {
									break;
								}
							} else {
								break;
							}
							
						} while ( parens.count() );
						
						if ( parens.isEmpty() )
						{
							bMatch = true;
							m.type = PMatch::Mismatch;
							break;
						}
					}
				}
				
				if ( bInsert )
				{
					parens.push(parenId);
					columns.push(offset);
					lengths.push(length);
					type.push(true);
					lines.push(line);
				}
			} else if ( close ) {
				lines.push(line);
				type.push(false);
				columns.push(offset);
				parens.push(parenId);
				lengths.push(length);
			} else if ( open ) {
				if ( (bMatch = (parens.top() == parenId) ) )
				{
					bMatch &= parens.count() == 1;
					
					if ( bMatch )
					{
						m.type = PMatch::Match;
						
						break;
					} else if ( parens.count() ) {
						parens.pop();
						lines.pop();
						type.pop();
						lengths.pop();
						columns.pop();
					} else
						qWarning("bad paren nesting...");
				} else {
					bMatch = true;
					
					parens.pop();
					type.pop();
					m.line[0] = lines.pop();
					m.column[0] = columns.pop();
					m.length[0] = lengths.pop();
					
					m.type = PMatch::Mismatch;
					
					break;
				}
			}
		}
		
		if ( bMatch )
			break;
		
		}
		
		--line;
		
		QDocumentLine b = d->line(line);
		pos = b.length();
		
		if ( !b.isValid() )
		{
			m.type = PMatch::Invalid;
			return;
		}
		
		tokens = b.tokens();
	}
	
	m.line[1] = line;
	m.column[1] = offset;
	m.length[1] = length;
}

/*!
	\brief Return the indent to use when inserting a line at a given cursor position
*/
QString QNFADefinition::indent(const QDocumentCursor& c)
{
	if ( c.isNull() || c.line().isNull() )
		return QString();
	
	QDocumentLine b = c.line();
	int pos, max = qMin(c.columnNumber(), b.text().size());
	
	QString s = b.text().left(max);
	
	//qDebug("line %i, column %i : %s", b.lineNumber(), c.columnNumber(), qPrintable(s));
	
	for ( pos = 0; pos < max; ++pos )
		if ( !s.at(pos).isSpace() )
			break;
	
	int indent = 0;
	bool open = false;
	QString spaces = s.left(pos);
	
	//qDebug("indent 
	
	QVector<int> tokens = b.tokens();
	
	if ( tokens.count() >= 4 )
	{
		int tokenIndex = 1;
		int tokenCount = tokens.first();
		
		while ( tokenIndex + 2 < tokenCount )
		{
			int offset = tokens.at(tokenIndex);
			int length = tokens.at(++tokenIndex);
			int tokenId = tokens.at(++tokenIndex);
			++tokenIndex;
			
			if ( offset >= c.columnNumber() )
				break;
			
			if ( !(tokenId & QLanguageDefinition::AutoIndent) )
				continue;
			
			bool popen = tokenId & QLanguageDefinition::OpenParen;
			bool pclose = tokenId & QLanguageDefinition::CloseParen;
			
			if ( popen )
				++indent;
			else if ( pclose )
				--indent;
		}
	}
	
	if ( indent > 0 )
		spaces += QString(indent, '\t');
	
	return spaces;
}

/*!
	\brief Determines whether the given key event at the given position should cause unindent to happen
*/
bool QNFADefinition::unindent (const QDocumentCursor& c, const QString& ktxt)
{
	if ( c.isNull() || c.line().isNull() || ktxt.isEmpty() )
		return false;
	
	QString s = c.line().text();
	int pos, max = qMin(c.columnNumber(), s.size());
	
	s.insert(max, ktxt);
	
	//qDebug("outdenting %s", qPrintable(s));
	for ( pos = 0; pos < max; ++pos ) 
		if ( !s.at(pos).isSpace() )
			break;
	
	if ( !pos || pos >= c.columnNumber() + ktxt.length() )
		return false;
	
	QString text = s.mid(pos);
	
	int flags = blockFlags(c.document(), c.lineNumber(), 0);
	
	// <trick>
	QNFA *context;
	QStack<QNFA*> cxt;
	QStack<int> cxtidx;
	QMap<int, QPair<int, int> > cxtData;
	
	cxt.push(m_root);
	
	if ( c.lineNumber() > 0 )
	{
		int min = 0;
		const QVector<int>& cxts = c.document()->line(c.lineNumber() - 1).contexts();
		
		for ( int i = min; i + 2 < cxts.count(); i += 3 )
		{
			// at this point we're only interested in context going forth
			if ( cxts.at(i + 1) != -1 )
				break;
			
			int cxtId = cxts.at(i + 2);
			
			// language are just there for proper context awareness, they
			// serve no actual purpose in the context stack of the lexer...
			if ( cxtId & Language )
				continue;
			
			//cxt.push(m_contextNames.value(contextName(cxtId)));
			cxt.push(m_contexts.value(cxtId).nfa);
		}
	}
	
	context = cxt.pop();
	
	QDocumentLineHandle *h = new QDocumentLineHandle(text);
	
	tokenize(h, context, cxt, cxtidx, cxtData);
	
	QVector<int> tokens = h->tokens();
	
	delete h;
	// </trick>
	
	if ( tokens.count() < 4 )
		return false;
	
	int tokenIndex = 1;
	int tokenCount = tokens.first();
	
	// only check the first token...
	//while ( tokenIndex + 2 < tokenCount )
	{
		int offset = tokens.at(tokenIndex);
		int length = tokens.at(++tokenIndex);
		int tokenId = tokens.at(++tokenIndex);
		++tokenIndex;
		
		bool popen = tokenId & QLanguageDefinition::OpenParen;
		bool pclose = tokenId & QLanguageDefinition::CloseParen;
		
		if ( !pclose || !(tokenId & QLanguageDefinition::Paren) )
			return false;
		
		//int parenId = m_tokens.at(tokenId & QLanguageDefinition::TokenIdMask).parenId;
		
		if ( 
				(tokenId & QLanguageDefinition::AutoIndent)
			&&
				(offset <= c.columnNumber() - pos)
			&&
				(offset + length > c.columnNumber() - pos)
			)
			return true;
		
	}
	
	return false;
}

int QNFADefinition::findBlockEnd(QDocument *d, int line, bool *open)
{
	QDocumentLine b = d->line(line);
	
	QVector<int> pidx, tokens = b.tokens();
	
	if ( !b.isValid() || tokens.count() < 4 )
		return -1;
	
	int tokenIndex = 1;
	int tokenCount = tokens.first();
	
	while ( tokenIndex + 2 < tokenCount )
	{
		int offset = tokens.at(tokenIndex);
		int length = tokens.at(++tokenIndex);
		int tokenId = tokens.at(++tokenIndex);
		++tokenIndex;
		
		if ( !(tokenId & QLanguageDefinition::Paren) )
			continue;
		
		bool popen = tokenId & QLanguageDefinition::OpenParen;
		bool pclose = tokenId & QLanguageDefinition::CloseParen;
		
		if ( !(popen || pclose) || !(tokenId & QLanguageDefinition::BlockFold) )
			continue;
		
		int parenId = m_tokens.at(tokenId & QLanguageDefinition::TokenIdMask).parenId;
		
		if ( pclose && pidx.count() )
			pidx.pop_back();
		
		if ( popen )
			pidx << tokenIndex - 3;
		
	}
	
	if ( pidx.count() )
	{
		int firstPidx = pidx.first();
		
		PMatch m;
		m.line[0] = line;
		m.column[0] = tokens.at(firstPidx);
		m.length[0] = tokens.at(firstPidx + 1);
		
		matchOpen(d, m);
		
		if (
				(m.type == PMatch::Match)
			||
				(
					(m.type == PMatch::Mismatch)
				&&
					(m.line[0] == line)
				&&
					(m.column[0] == tokens.at(firstPidx))
				)
			)
		{
			if ( open )
			{
				int flags = blockFlags(d, m.line[1], 0);
				
				*open = QCE_FOLD_OPEN_COUNT(flags) | (m.type == PMatch::Mismatch);
			}
			
			return m.line[1];
		} else if ( m.type == PMatch::Invalid ) {
			return d->lines() - 1;
		}
	}
	
	return -1;
}

/*!
	\brief Expand a collapsed block at a given line
*/
void QNFADefinition::expand(QDocument *d, int line)
{
	QDocumentLine b = d->line(line);
	
	if ( !b.isValid() || !b.hasFlag(QDocumentLine::CollapsedBlockStart) )
		return;
	
	bool open = false;
	int end = findBlockEnd(d, line, &open);
	
	int count = end - line;
	
	if ( open )
		--count;
	
	if ( count <= 0 )
		return;
	
	b.setFlag(QDocumentLine::CollapsedBlockStart, false);
	
	bool shared = false;
	int i = line + 1;
	
	while ( i < end )
	{
		QDocumentLine l = d->line(i);
		l.setFlag(QDocumentLine::Hidden, false);
		//qDebug("+%i", i);
		if ( l.hasFlag(QDocumentLine::CollapsedBlockStart) )
		{
			bool bopen = false;
			int bend = findBlockEnd(d, i, &bopen);
			
			i = bend;

			if ( !bopen )
				++i;
			
			if ( bend == end )
			{
				shared = true;
			}
		} else {
			++i;
		}
	}
	
	if ( !shared )
	{
		d->line(end).setFlag(QDocumentLine::CollapsedBlockEnd, false);
	}
	
	//if ( !open && !shared )
	{
		//qDebug("+%i", end);
		d->line(end).setFlag(QDocumentLine::Hidden, false);
	}
	
	//qDebug("expanding %i lines from %i", count, line);
	d->impl()->showEvent(line, count);
}

/*!
	\brief Collapse a text block at a given line
*/
void QNFADefinition::collapse(QDocument *d, int line)
{
	QDocumentLine b = d->line(line);
	
	if ( !b.isValid() || b.hasFlag(QDocumentLine::CollapsedBlockStart) )
		return;
	
	//qDebug("collapse line %i", line);
	
	bool open = false;
	int end = findBlockEnd(d, line, &open);
	
	int count = end - line;
	
	if ( open )
		--count;
	
	if ( count <= 0 )
		return;
	
	b.setFlag(QDocumentLine::CollapsedBlockStart);
	
	for ( int i = line + 1; i < end; ++i )
	{
		//if ( !d->line(i).hasFlag(QDocumentLine::Hidden) )
		//{
			//qDebug("-%i", i);
			d->line(i).setFlag(QDocumentLine::Hidden);
		//}
	}
	
	d->line(end).setFlag(QDocumentLine::CollapsedBlockEnd);
	
	if ( !open )
	{
		if ( !d->line(end).hasFlag(QDocumentLine::Hidden) )
		{
			//qDebug("-%i", end);
			d->line(end).setFlag(QDocumentLine::Hidden);
		}
	}
	
	//qDebug("collapsing %i lines from %i", count, line);
	d->impl()->hideEvent(line, count);
	
	#if 0
	if ( !m_indentFold )
		return;
	
	int i = ln + 1, k;
	QDocumentLine block = d->line(i);
	
	if ( !block.isValid() )
		return;
	
	indent = block.firstChar();
	
	do
	{
		block = d->line(++i);
		k = block.firstChar();
	} while ( block.isValid() && ((k >= indent) || (k == -1)) );
	
	if ( k < indent )
	{
		block = d->line(--i);
		k = block.firstChar();
	}
	
	while ( block.isValid() && (i > ln) && (k == -1) )
	{
		block = d->line(--i);
		k = block.firstChar();
	}
	
	b.setFlag(QDocumentLine::CollapsedBlockStart);
	
	//qDebug("hidding from %i to %i", ln + 1, i);
	
	for ( int j = ln + 1; j <= i; ++j )
	{
		d->line(j).setFlag(QDocumentLine::Hidden);
	}
	
	d->line(i).setFlag(QDocumentLine::CollapsedBlockEnd);
	
	d->impl()->hideEvent(ln, i - ln);
	#endif
}

/*!
	\brief Compute the collapse state of a line
*/
int QNFADefinition::blockFlags(QDocument *d, int line, int depth) const
{
	QDocumentLine b = d->line(line);
	
	int flags = blockFlags(b.tokens());
	
	if ( b.hasFlag(QDocumentLine::CollapsedBlockStart) )
		flags |= Collapsed;
	
	return flags;
}

int QNFADefinition::blockFlags(const QVector<int>& tokens) const
{
	if ( tokens.count() < 4 )
		return None;
	
	int ret = None;
	short open = 0, close = 0;
	
	int tokenIndex = 1;
	int tokenCount = tokens.first();
	
	while ( tokenIndex + 2 < tokenCount )
	{
		int offset = tokens.at(tokenIndex);
		int length = tokens.at(++tokenIndex);
		int tokenId = tokens.at(++tokenIndex);
		++tokenIndex;
		
		bool popen = tokenId & QLanguageDefinition::OpenParen;
		bool pclose = tokenId & QLanguageDefinition::CloseParen;
		
		if ( !(popen || pclose) || !(tokenId & QLanguageDefinition::BlockFold) )
			continue;
		
		int parenId = m_tokens.at(tokenId & QLanguageDefinition::TokenIdMask).parenId;
		
		if ( pclose )
		{
			if ( open )
				--open;
			else
				++close;
			
		}
		
		if ( popen )
		{
			++open;
		}
	}
	
	if ( open )
		ret |= Collapsible;
	
	if ( close )
		ret |= Closure;
	
	//qDebug("%i : ret : %i (%i, %i)", line, ret, ret & 0xffff0000, short(((short)0) | (ret & 0xffff)));
	
	return QCE_FOLD_FLAGS(ret, open, close);
}

int& QNFADefinition::addContext(const QString& id, QNFA *nfa)
{
	//qDebug("registering context : %s", qPrintable(id));
	
	int cid = QLanguageDefinition::addContext(id);
	
	Context cxt;
	cxt.id = cid;
	cxt.d = this;
	cxt.nfa = nfa;
	m_contexts[cid] = cxt;
	m_contextIds[nfa] = cid;
	m_contextNames[id] = nfa;
	
	return m_contexts[cid].format;
}

void QNFADefinition::flushEmbedRequests(const QString& lang)
{
	//qDebug("flushing requests for : %s", qPrintable(lang));
	QHash<QString, EmbedRequestList>::iterator it;
	
	it = m_pendingEmbeds.begin();
	
	while ( it != m_pendingEmbeds.end() )
	{
		QString r = it.key();
		
		if (
				r.startsWith(lang)
			&&
				(
					(r.count() == lang.count())
				||
					(r.at(lang.count()) == ':')
				)
			)
		{
			QNFA *src = m_contextNames.value(r);
			
			if ( !src )
			{
				++it;
				continue;
			}
			
			foreach( const EmbedRequest& request, *it )
			{
				//qDebug("embedding %s in 0x%x at index %i",
				//		qPrintable(r), request.target, request.index);
				
				embed(src, request.target, request.index);
			}
			
			it = m_pendingEmbeds.erase(it);
		} else {
			++it;
		}
	}
}

void QNFADefinition::addEmbedRequest(const QString& cxt, QNFA *dest)
{
	//qDebug("Adding request : %s", qPrintable(cxt));
	
	if ( m_contextNames.contains(cxt) )
	{
		//qDebug("dealing with request : %s", qPrintable(cxt));
		embed(m_contextNames[cxt], dest, dest->out.branch->count());
	} else {
		m_pendingEmbeds[cxt] << EmbedRequest(dest, dest->out.branch->count());
	}
}

void QNFADefinition::shareEmbedRequests(QNFA *src, QNFA *dest, int offset)
{
	QHash<QString, EmbedRequestList>::iterator it = m_pendingEmbeds.begin();
	
	while ( it != m_pendingEmbeds.end() )
	{
		foreach ( const EmbedRequest& request, *it )
		{
			if ( request.target == src )
			{
				//qDebug("sharing one embed request between 0x%x and 0x%x at %i + %i", src, dest, offset, request.index);
				it->append(EmbedRequest(dest, request.index + offset));
			}
		}
		
		++it;
	}
}
